@@ -40,13 +40,20 @@
 #include <sys/ioctl.h>
 
 #include <pciaccess.h>
+
+#include <xorg-server.h>
 #include <xf86.h>
 #include <xf86drm.h>
 #include <xf86drmMode.h>
 #include <xf86_OSproc.h>
 #include <i915_drm.h>
 
+#ifdef XSERVER_PLATFORM_BUS
+#include <xf86platformBus.h>
+#endif
+
 #include "intel_driver.h"
+#include "fd.h"
 
 struct intel_device {
 	char *master_node;
@@ -58,6 +65,32 @@ struct intel_device {
 
 static int intel_device_key = -1;
 
+static int __intel_get_device_id(int fd)
+{
+	struct drm_i915_getparam gp;
+	int devid = 0;
+
+	memset(&gp, 0, sizeof(gp));
+	gp.param = I915_PARAM_CHIPSET_ID;
+	gp.value = &devid;
+
+	if (ioctl(fd, DRM_IOCTL_I915_GETPARAM, &gp, sizeof(gp)))
+		return 0;
+
+	return devid;
+}
+
+int intel_entity_get_devid(int idx)
+{
+	struct intel_device *dev;
+
+	dev = xf86GetEntityPrivate(idx, intel_device_key)->ptr;
+	if (dev == NULL)
+		return 0;
+
+	return __intel_get_device_id(dev->fd);
+}
+
 static inline struct intel_device *intel_device(ScrnInfoPtr scrn)
 {
 	if (scrn->entityList == NULL)
@@ -110,40 +143,6 @@ static int __intel_check_device(int fd)
 	return ret;
 }
 
-static int fd_set_cloexec(int fd)
-{
-	int flags;
-
-	if (fd == -1)
-		return fd;
-
-#ifdef FD_CLOEXEC
-	flags = fcntl(fd, F_GETFD);
-	if (flags != -1) {
-		flags |= FD_CLOEXEC;
-		fcntl(fd, F_SETFD, flags);
-	}
-#endif
-
-	return fd;
-}
-
-static int fd_set_nonblock(int fd)
-{
-	int flags;
-
-	if (fd == -1)
-		return fd;
-
-	flags = fcntl(fd, F_GETFD);
-	if (flags != -1) {
-		flags |= O_NONBLOCK;
-		fcntl(fd, F_SETFD, flags);
-	}
-
-	return fd;
-}
-
 static int __intel_open_device(const struct pci_device *pci, char **path)
 {
 	int fd;
@@ -152,6 +151,9 @@ static int __intel_open_device(const struct pci_device *pci, char **path)
 		char id[20];
 		int ret;
 
+		if (pci == NULL)
+			return -1;
+
 		snprintf(id, sizeof(id),
 			 "pci:%04x:%02x:%02x.%d",
 			 pci->domain, pci->bus, pci->dev, pci->func);
@@ -188,6 +190,27 @@ static int __intel_open_device(const struct pci_device *pci, char **path)
 	return fd;
 }
 
+static char *find_master_node(int fd)
+{
+	struct stat st, master;
+	char buf[128];
+
+	if (fstat(fd, &st))
+		return NULL;
+
+	if (!S_ISCHR(st.st_mode))
+		return NULL;
+
+	sprintf(buf, "/dev/dri/card%d", (int)(st.st_rdev & 0x7f));
+	if (stat(buf, &master) == 0 &&
+	    st.st_mode == master.st_mode &&
+	    (st.st_rdev & 0x7f) == master.st_rdev)
+		return strdup(buf);
+
+	/* Fallback to iterating over the usual suspects */
+	return drmGetDeviceNameFromFd(fd);
+}
+
 static char *find_render_node(int fd)
 {
 #if defined(USE_RENDERNODE)
@@ -204,23 +227,64 @@ static char *find_render_node(int fd)
 	if (master.st_rdev & 0x80)
 		return NULL;
 
-	sprintf(buf, "/dev/dri/renderD%d", (int)((master.st_rdev | 0x80) & 0xff));
+	sprintf(buf, "/dev/dri/renderD%d", (int)((master.st_rdev | 0x80) & 0xbf));
 	if (stat(buf, &render) == 0 &&
 	    master.st_mode == render.st_mode &&
-	    render.st_rdev == (master.st_rdev | 0x80))
+	    render.st_rdev == ((master.st_rdev | 0x80) & 0xbf))
 		return strdup(buf);
 #endif
 
 	return NULL;
 }
 
+#if defined(ODEV_ATTRIB_PATH)
+static char *get_path(struct xf86_platform_device *dev)
+{
+	const char *path;
+
+	if (dev == NULL)
+		return NULL;
+
+	path = xf86_get_platform_device_attrib(dev, ODEV_ATTRIB_PATH);
+	if (path == NULL)
+		return NULL;
+
+	return strdup(path);
+}
+
+#else
+
+static char *get_path(struct xf86_platform_device *dev)
+{
+	return NULL;
+}
+#endif
+
+
+#if defined(ODEV_ATTRIB_FD)
+static int get_fd(struct xf86_platform_device *dev)
+{
+	if (dev == NULL)
+		return -1;
+
+	return xf86_get_platform_device_int_attrib(dev, ODEV_ATTRIB_FD, -1);
+}
+
+#else
+
+static int get_fd(struct xf86_platform_device *dev)
+{
+	return -1;
+}
+#endif
+
 int intel_open_device(int entity_num,
 		      const struct pci_device *pci,
-		      const char *path)
+		      struct xf86_platform_device *platform)
 {
 	struct intel_device *dev;
 	char *local_path;
-	int fd;
+	int fd, master_count;
 
 	if (intel_device_key == -1)
 		intel_device_key = xf86AllocateEntityPrivateIndex();
@@ -231,11 +295,23 @@ int intel_open_device(int entity_num,
 	if (dev)
 		return dev->fd;
 
-	local_path = path ? strdup(path) : NULL;
+	local_path = get_path(platform);
 
-	fd = __intel_open_device(pci, &local_path);
-	if (fd == -1)
-		goto err_path;
+	master_count = 1; /* DRM_MASTER is managed by Xserver */
+	fd = get_fd(platform);
+	if (fd == -1) {
+		fd = __intel_open_device(pci, &local_path);
+		if (fd == -1)
+			goto err_path;
+
+		master_count = 0;
+	}
+
+	if (local_path == NULL) {
+		local_path = find_master_node(fd);
+		if (local_path == NULL)
+			goto err_close;
+	}
 
 	if (!__intel_check_device(fd))
 		goto err_close;
@@ -246,7 +322,7 @@ int intel_open_device(int entity_num,
 
 	dev->fd = fd;
 	dev->open_count = 0;
-	dev->master_count = 0;
+	dev->master_count = master_count;
 	dev->master_node = local_path;
 	dev->render_node = find_render_node(fd);
 	if (dev->render_node == NULL)
@@ -263,7 +339,8 @@ int intel_open_device(int entity_num,
 	return fd;
 
 err_close:
-	close(fd);
+	if (master_count == 0) /* Don't close server-fds */
+		close(fd);
 err_path:
 	free(local_path);
 	return -1;
@@ -317,6 +394,13 @@ const char *intel_get_client_name(ScrnInfoPtr scrn)
 	return dev->render_node;
 }
 
+int intel_get_device_id(ScrnInfoPtr scrn)
+{
+	struct intel_device *dev = intel_device(scrn);
+	assert(dev && dev->fd != -1);
+	return __intel_get_device_id(dev->fd);
+}
+
 int intel_get_master(ScrnInfoPtr scrn)
 {
 	struct intel_device *dev = intel_device(scrn);
@@ -358,20 +442,6 @@ int intel_put_master(ScrnInfoPtr scrn)
 	return ret;
 }
 
-void __intel_uxa_release_device(ScrnInfoPtr scrn)
-{
-	struct intel_device *dev = intel_device(scrn);
-	if (dev && dev->open_count == 0) {
-		intel_set_device(scrn, NULL);
-
-		drmClose(dev->fd);
-		if (dev->render_node != dev->master_node)
-			free(dev->render_node);
-		free(dev->master_node);
-		free(dev);
-	}
-}
-
 void intel_put_device(ScrnInfoPtr scrn)
 {
 	struct intel_device *dev = intel_device(scrn);
