@@ -17,7 +17,7 @@ static inline bool need_redirect(struct sna *sna, PixmapPtr dst)
 		dst->drawable.height > sna->render.max_3d_size);
 }
 
-static inline float pack_2s(int16_t x, int16_t y)
+static force_inline float pack_2s(int16_t x, int16_t y)
 {
 	union {
 		struct sna_coordinate p;
@@ -28,35 +28,43 @@ static inline float pack_2s(int16_t x, int16_t y)
 	return u.f;
 }
 
-static inline int vertex_space(struct sna *sna)
+static force_inline int vertex_space(struct sna *sna)
 {
 	return sna->render.vertex_size - sna->render.vertex_used;
 }
-static inline void vertex_emit(struct sna *sna, float v)
+static force_inline void vertex_emit(struct sna *sna, float v)
 {
 	assert(sna->render.vertex_used < sna->render.vertex_size);
 	sna->render.vertices[sna->render.vertex_used++] = v;
 }
-static inline void vertex_emit_2s(struct sna *sna, int16_t x, int16_t y)
+static force_inline void vertex_emit_2s(struct sna *sna, int16_t x, int16_t y)
 {
 	vertex_emit(sna, pack_2s(x, y));
 }
 
-static inline int batch_space(struct sna *sna)
+static force_inline int batch_space(struct sna *sna)
 {
 	assert(sna->kgem.nbatch <= KGEM_BATCH_SIZE(&sna->kgem));
 	assert(sna->kgem.nbatch + KGEM_BATCH_RESERVED <= sna->kgem.surface);
 	return sna->kgem.surface - sna->kgem.nbatch - KGEM_BATCH_RESERVED;
 }
 
-static inline void batch_emit(struct sna *sna, uint32_t dword)
+static force_inline void batch_emit(struct sna *sna, uint32_t dword)
 {
 	assert(sna->kgem.mode != KGEM_NONE);
 	assert(sna->kgem.nbatch + KGEM_BATCH_RESERVED < sna->kgem.surface);
 	sna->kgem.batch[sna->kgem.nbatch++] = dword;
 }
 
-static inline void batch_emit_float(struct sna *sna, float f)
+static force_inline void batch_emit64(struct sna *sna, uint64_t qword)
+{
+	assert(sna->kgem.mode != KGEM_NONE);
+	assert(sna->kgem.nbatch + 2 + KGEM_BATCH_RESERVED < sna->kgem.surface);
+	*(uint64_t *)(sna->kgem.batch+sna->kgem.nbatch) = qword;
+	sna->kgem.nbatch += 2;
+}
+
+static force_inline void batch_emit_float(struct sna *sna, float f)
 {
 	union {
 		uint32_t dw;
@@ -105,15 +113,37 @@ static inline bool
 unattached(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
-	return priv == NULL || (priv->gpu_damage == NULL && priv->cpu_damage);
+	return priv == NULL || (priv->gpu_damage == NULL && priv->cpu_damage && !priv->cpu_bo);
 }
 
 static inline bool
-picture_is_gpu(struct sna *sna, PicturePtr picture)
+picture_is_gpu(struct sna *sna, PicturePtr picture, unsigned flags)
 {
-	if (!picture || !picture->pDrawable)
+	if (!picture)
 		return false;
-	return is_gpu(sna, picture->pDrawable, PREFER_GPU_RENDER);
+
+	if (!picture->pDrawable) {
+		switch (flags) {
+		case PREFER_GPU_RENDER:
+			switch (picture->pSourcePict->type) {
+			case SourcePictTypeSolidFill:
+			case SourcePictTypeLinear:
+				return false;
+			default:
+				return true;
+			}
+		case PREFER_GPU_SPANS:
+			return true;
+		default:
+			return false;
+		}
+	} else {
+		if (picture->repeat &&
+		    (picture->pDrawable->width | picture->pDrawable->height) == 1)
+			return flags == PREFER_GPU_SPANS;
+	}
+
+	return is_gpu(sna, picture->pDrawable, flags);
 }
 
 static inline bool
